from .base import BaseSegmenter

class FigureSegmenter(BaseSegmenter):
    def __init__(
        self,
        margin=15,
        overlap_margin=2,
        caption_margin=30,
        min_dimension=20,
        caption_keywords=["figure", "fig"],
        fig_scale=4,
    ):
        super().__init__()

        self.margin = margin
        self.caption_keywords = caption_keywords
        self.min_dimension = min_dimension
        self.fig_scale = fig_scale
        self.overlap_margin = overlap_margin
        self.caption_margin = caption_margin

    def get_image_segments(self, page, blocks):
        """Extract image segments from the page."""
        image_segments = []

        for block in blocks:
            if "image" in block:
                bbox = block["bbox"]
                image_segments.append(self.new_segment(bbox, page=page, image="png"))

        return image_segments

    def get_caption_segments(self, page, blocks):
        """Extract caption segments from the page."""
        caption_segments = []

        for block in blocks:
            if "lines" in block:

                combined_text = []
                block_bboxes = []
                for line in block["lines"]:
                    for span in line["spans"]:
                        combined_text.append(span["text"])
                        block_bboxes.append(span["bbox"])

                # Combine all text into a paragraph
                paragraph_text = " ".join(combined_text)

                bbox = self.merge_bounding_boxes(block_bboxes)

                if any(
                    [
                        paragraph_text[:15].lower().startswith(keyword)
                        for keyword in self.caption_keywords
                    ]
                ):
                    
                    caption_segments.append(
                        self.new_segment(bbox, page=page, text=paragraph_text)
                    )
                    

        return caption_segments

    def get_text_segments(self, page, blocks):
        """Extract text segments from the page."""
        text_segments = []

        for block in blocks:
            if "lines" in block:

                combined_text = []
                block_bboxes = []
                for line in block["lines"]:
                    for span in line["spans"]:
                        combined_text.append(span["text"])
                        block_bboxes.append(span["bbox"])

                # Combine all text into a paragraph
                paragraph_text = " ".join(combined_text)

                bbox = self.merge_bounding_boxes(block_bboxes)

                text_segments.append(
                    self.new_segment(bbox, page=page, text=paragraph_text)
                )

        return text_segments

    def get_shape_segments(self, page):
        """Extract shape segments from the page."""
        shapes = page.get_drawings()
        shape_segments = []

        for shape in shapes:
            bbox = shape["rect"]
            shape_segments.append(self.new_segment(bbox, page=page))

        return shape_segments

    def merge_segments(
        self, page, image_segments, caption_segments, shape_segments, text_segments
    ):
        """Try to combine image and caption segments into figure segments.

        This function tries to merge image and catption segments.
        Image segments are first combined with possible overlapping image and shape segments.
        If a caption is close to an image segment, they are combined also.
        """

        # first group image and shape segments
        graphical_segments = image_segments + shape_segments

        # order the graphical segments by size
        graphical_segments = sorted(
            graphical_segments,
            key=lambda x: (
                x["position"]["boundingRect"]["x2"]
                - x["position"]["boundingRect"]["x1"]
            )
            * (
                x["position"]["boundingRect"]["y2"]
                - x["position"]["boundingRect"]["y1"]
            ),
            reverse=True,
        )

        # condense the bounding boxes, by iterating over the graphical segments. If they overlap within
        # the margin, they are combined.
        found_overlap = True
        while found_overlap:
            found_overlap = False
            for i, segment in enumerate(graphical_segments):
                for j, other_segment in enumerate(graphical_segments):
                    if i != j:
                        bbox = segment["position"]["boundingRect"]
                        other_bbox = other_segment["position"]["boundingRect"]
                        if (
                            bbox["x1"] - self.margin < other_bbox["x2"]
                            and bbox["x2"] + self.margin > other_bbox["x1"]
                            and bbox["y1"] - self.margin < other_bbox["y2"]
                            and bbox["y2"] + self.margin > other_bbox["y1"]
                        ):
                            combined_segment = self.combine_segments(
                                segment, other_segment
                            )
                            graphical_segments[i] = combined_segment
                            graphical_segments.pop(j)
                            found_overlap = True
                            break
                if found_overlap:
                    break

        # remove segments that are too small
        def segment_bigh_enough(segment):
            bbox = segment["position"]["boundingRect"]
            delta_x = bbox["x2"] - bbox["x1"]
            delta_y = bbox["y2"] - bbox["y1"]
            return delta_x > self.min_dimension and delta_y > self.min_dimension

        graphical_segments = [
            segment for segment in graphical_segments if segment_bigh_enough(segment)
        ]

        # if there are text_segments that overlap with graphical segments, extend the bounding box
        for segment in graphical_segments:
            bbox = segment["position"]["boundingRect"]
            for text_segment in text_segments:
                text_bbox = text_segment["position"]["boundingRect"]
                if (
                    bbox["x1"] - self.overlap_margin < text_bbox["x2"]
                    and bbox["x2"] + self.overlap_margin > text_bbox["x1"]
                    and bbox["y1"] - self.overlap_margin < text_bbox["y2"]
                    and bbox["y2"] + self.overlap_margin > text_bbox["y1"]
                ):
                    combined_bbox = self.merge_pdfjs_bounding_boxes([bbox, text_bbox])
                    segment["position"]["boundingRect"] = combined_bbox
                    # add the text to the graphical segment
                    if "text" not in segment["content"]:
                        segment["content"]["text"] = text_segment["content"]["text"]
                    else:
                        segment["content"]["text"] += (
                            "\n" + text_segment["content"]["text"]
                        )

        # now try to combine the graphical segments with the caption segments
        closest_caption_segment_index = [None] * len(graphical_segments)
        for segment in graphical_segments:

            graphical_bbox = segment["position"]["boundingRect"]

            # find the closest graphical segment
            closest_segment = None
            closest_distance = float("inf")
            closest_index = None
            for i, caption_segment in enumerate(caption_segments):
                bbox = caption_segment["position"]["boundingRect"]
                # use a minimum distance of the caption margin
                distance = min(
                    min(
                        abs(bbox["x1"] - graphical_bbox["x2"]),
                        abs(bbox["x2"] - graphical_bbox["x1"]),
                    ),
                    min(
                        abs(bbox["y1"] - graphical_bbox["y2"]),
                        abs(bbox["y2"] - graphical_bbox["y1"]),
                    ),
                )
                if distance < closest_distance:
                    closest_distance = distance
                    closest_segment = caption_segment
                    closest_index = i

            # if a closest segment is found and it is close enough, combine the caption with the graphical segment
            if closest_segment is None:
                continue

            closest_caption_segment_index[graphical_segments.index(segment)] = (
                closest_index
            )

        # if there are graphical segments with same closest_caption_segment_index, combine them
        for i, segment in enumerate(graphical_segments):
            closest_index = closest_caption_segment_index[i]
            if closest_index is not None:
                # update the caption text accordingly
                if "text" not in segment["content"]:
                    segment["content"]["text"] = caption_segments[closest_index][
                        "content"
                    ]["text"]
                else:
                    segment["content"]["text"] = (
                        caption_segments[closest_index]["content"]["text"]
                        + "\n\n" + segment["content"]["text"]
                    )
                # add the rects
                if "rects" in caption_segments[closest_index]["position"]:
                    bbrect = caption_segments[closest_index]["position"]["boundingRect"]
                    bbrect["extra"] = "caption"
                    segment["position"]["rects"].append(bbrect)

                for j, other_segment in enumerate(graphical_segments):
                    if i != j and closest_caption_segment_index[j] == closest_index:
                        combined_segment = self.combine_segments(segment, other_segment)
                        graphical_segments[i] = combined_segment
                        graphical_segments.pop(j)
                        closest_caption_segment_index.pop(j)
                        break

        return graphical_segments

    def calculate_figure_segments(self, page, combined_segments):
        import pymupdf as fitz
        import base64

        def encode_image_to_base64(image_bytes):
            return base64.b64encode(image_bytes).decode("utf-8")

        def get_image_segment(segment):

            bbox = segment["position"]["boundingRect"]

            bbox = fitz.Rect(bbox["x1"], bbox["y1"], bbox["x2"], bbox["y2"])
            image = page.get_pixmap(
                matrix=fitz.Matrix(self.fig_scale, self.fig_scale), clip=bbox
            )

            # image is a Pixmap object, convert it to raw image bytes
            image_bytes = image.tobytes()

            # encode the image as base64
            image_base64 = encode_image_to_base64(image_bytes)

            segment["content"]["image"] = f"data:image/png;base64,{image_base64}"

            return segment

        return [get_image_segment(segment) for segment in combined_segments]

    def run_page(self, page):

        blocks = page.get_text("dict")["blocks"]

        image_segments = self.get_image_segments(page, blocks)
        caption_segments = self.get_caption_segments(page, blocks)
        shape_segments = self.get_shape_segments(page)
        text_segments = self.get_text_segments(page, blocks)

        combined_segments = self.merge_segments(
            page, image_segments, caption_segments, shape_segments, text_segments
        )
        figure_segments = self.calculate_figure_segments(page, combined_segments)

        return figure_segments

    def _run(self, doc, excluded_boxes):

        if len(excluded_boxes) > 0:
            raise NotImplementedError(
                "FigureSegmenter does not support excluded_boxes yet."
            )

        self.figure_segments = []

        for page in doc:
            self.figure_segments.extend(self.run_page(page))

        return self.figure_segments



from .base import BaseSegmenter
from .sentence import SentenceSegmenter
from .mongo import MongoDBConnection


class SemanticSegmenter(BaseSegmenter):
    def __init__(self, max_char_joined_segments=1000, max_joined_segments=50):
        super().__init__()

        self.sentence_segmenter = SentenceSegmenter()
        self.max_char_joined_segments = max_char_joined_segments
        self.max_joined_segments = max_joined_segments

    def get_similarity_matrix(self, segments):
        import numpy as np

        max_length = self.max_char_joined_segments
        band = self.max_joined_segments

        similarity_matrix = np.zeros((len(segments), len(segments)))

        embedding_vectors = [np.array(el["embedding"]) for el in segments]

        segment_lengths = [len(el["content"]["text"]) for el in segments]

        for i in range(len(segments)):
            similarity_matrix[i][i] = 1.0
            for j in range(i+1, min(len(segments), i+band)):
                length_sum = np.sum(segment_lengths[i:j+1])

                if length_sum > max_length:
                    break

                similarity = np.dot(embedding_vectors[i], embedding_vectors[j]) / (
                    np.linalg.norm(embedding_vectors[i]) * np.linalg.norm(embedding_vectors[j]))
                similarity_matrix[i][j] = similarity
                similarity_matrix[j][i] = similarity

        return similarity_matrix

    def find_best_segmentation(self, similarity):
        import numpy as np

        if isinstance(similarity, list):
            similarity = np.array(similarity)

        # neglect non calculated values
        similarity[similarity == 0] = np.nan

        # calculate the running median of the similarity matrix on a 20x20 window
        def running_quantile(data, window_size=40, quantile=0.4):
            data = np.array(data)
            data = np.pad(data, (window_size//2, window_size//2),
                          mode='constant', constant_values=np.nan)
            data = np.array([np.nanquantile(data[i-window_size//2:i+window_size//2+1], quantile)
                            for i in range(window_size//2, len(data)-window_size//2)])
            return data
        median_similiarty = running_quantile(similarity)

        # use the median as a threshold for possible grouping
        matching_similarity = similarity > median_similiarty

        # find maximale blocks
        def find_max_blocks(matching_similarity, window_size=50):
            blocks = []
            for i in range(len(matching_similarity)):
                for j in range(i, len(matching_similarity)):
                    if any(~matching_similarity[i:j, i:j].flatten()):
                        blocks.append(j-i-1)
                        break

            blocks.append(1)
            blocks.append(1)

            return blocks
        blocks = find_max_blocks(matching_similarity)

        # generate the first join recipy with maximal sized blocks
        to_join = []

        while max(blocks) > 0:
            current_max = max(blocks)

            # find all matching indices
            idx = [i for i, el in enumerate(blocks) if el == current_max]

            # add the matching indices to the to_join array
            to_join.extend([list(range(el, el+current_max)) for el in idx])

            # make a set of the indices in to join
            found_indices = set(
                [item for sublist in to_join for item in sublist])

            # set the values in the blocks array to 0 for the indices in the to_join array
            for i in found_indices:
                blocks[i] = 0

        # sort the to_join array by the first element of each sublist
        to_join = sorted(to_join, key=lambda x: x[0])


        # some indices might apprear multiple times, we create a clever algorithm to remove them :D
        def remove_duplicate_indices(to_join):

            to_join = sorted(to_join, key=lambda x: x[0])

            # get a list of the indices that is found multiple times
            to_join_flat = [item for sublist in to_join for item in sublist]
            mutliple_indices = list(set([item for item in to_join_flat if to_join_flat.count(item) > 1]))

            # group the multiple indices in groups with consecutive indices
            multiple_groups = []
            for i in mutliple_indices:
                if not multiple_groups:
                    multiple_groups.append([i])
                else:
                    if i == multiple_groups[-1][-1] + 1:
                        multiple_groups[-1].append(i)
                    else:
                        multiple_groups.append([i])    

            found_elements_to_remove = False

            for el in multiple_groups:
                num_indices = len(el)
                lower_indices = el[0:num_indices//2]
                upper_indices = el[num_indices//2:]     

                if len(lower_indices) > 0:
                    # searchin the to_join array if a group starts with lower_indices, if yes, remove them
                    for i in range(len(to_join)):
                        if len(to_join[i]) < len(lower_indices):
                            continue

                        if to_join[i][0] == lower_indices[0]:
                            to_join[i] = to_join[i][len(lower_indices):]
                            found_elements_to_remove = True

                            

                if len(upper_indices) > 0:
                    # searchin the to_join array if a group ends with upper_indices, if yes, remove them
                    for i in range(len(to_join)):
                        if len(to_join[i]) < len(upper_indices):
                            continue

                        if to_join[i][-1] == upper_indices[-1]:
                            to_join[i] = to_join[i][:-len(upper_indices)]
                            found_elements_to_remove = True

            # remove empty lists
            to_join = [el for el in to_join if el]

            return to_join, found_elements_to_remove
        
        found_duplicates = True
        while found_duplicates:
            to_join, found_duplicates = remove_duplicate_indices(to_join)


        # add back indices that are not in the to_join array
        flat_to_join = [item for sublist in to_join for item in sublist]
        for i in range(len(similarity)):
            if i not in flat_to_join:
                to_join.append([i])

        return to_join


    def _run(self, doc, excluded_boxes):
        import numpy as np
        import json

        self.sentence_segments = self.sentence_segmenter._run(doc, excluded_boxes)

        # use embedding model to calculate for each sentence a vector
        for segment in self.sentence_segments:
            segment["embedding"] = self.get_embedding(
                segment["content"]["text"])

        # calculate a similarity matrix between all sentences
        similarity_matrix = self.get_similarity_matrix(self.sentence_segments)
        
        # find the best segmentation
        to_join = self.find_best_segmentation(similarity_matrix)

        # check if the to_join array can be right
        # 1. no duplicates:
        flat_to_join = [item for sublist in to_join for item in sublist]
        assert len(flat_to_join) == len(set(flat_to_join)), "Duplicates found in to_join array"

        # 2. all indices are in the range of the sentence_segments
        assert len(flat_to_join) == len(self.sentence_segments), "Not all indices are in the range of the sentence_segments"
        
        segments = []

        for join in to_join:
            if len(join) == 1:
                segments.append(self.sentence_segments[join[0]])
            else:
                new_segment = self.sentence_segments[join[0]]
                for idx in join[1:]:
                    new_segment = self.combine_segments(
                        new_segment, self.sentence_segments[idx]
                    )
                segments.append(new_segment)

        return segments
                    


from .mongo import MongoDBConnection



def get_embedding(text, mg, model="text-embedding-ada-002"):
    """
    Get the embedding of a text using the OpenAI API.
    
    """
    import openai

    res = mg.embeddings.find_one({"text": text}) or {}

    if res:
        return res["embedding"]

    response = openai.embeddings.create(
        input=text,
        model=model,
    )

    embedding = response.data[0].embedding

    mg.embeddings.insert_one({"text": text, "embedding": embedding})

    return embedding
    


class BaseSegmenter:
    def __init__(self):
        pass

    def convert_coordinates_from_pymupdf_to_pdfjs(
        self, bbox, page_height, page_width, page_number=None
    ):
        self.file_mode = ""
        self.file_id = ""

        output = {
            "x1": bbox[0],
            "y1": bbox[1],
            "x2": bbox[2],
            "y2": bbox[3],
            "width": page_width,
            "height": page_height,
        }
        if page_number is not None:
            output["pageNumber"] = page_number
        return output

    def merge_pdfjs_bounding_boxes(self, bboxes):
        """Calculate the union of multiple bounding boxes."""
        x1 = min(box["x1"] for box in bboxes)
        y1 = min(box["y1"] for box in bboxes)
        x2 = max(box["x2"] for box in bboxes)
        y2 = max(box["y2"] for box in bboxes)
        return {
            "x1": x1,
            "y1": y1,
            "x2": x2,
            "y2": y2,
            "width": bboxes[0]["width"],
            "height": bboxes[0]["height"],
        }

    def merge_bounding_boxes(self, bboxes):
        """Calculate the union of multiple bounding boxes."""
        x1 = min(box[0] for box in bboxes)
        y1 = min(box[1] for box in bboxes)
        x2 = max(box[2] for box in bboxes)
        y2 = max(box[3] for box in bboxes)
        return (x1, y1, x2, y2)


    def generate_id(self, segment):
        import hashlib
        import uuid
        if "content" in segment:
            if "text" in segment["content"]:
                content_hash = hashlib.sha256(segment["content"]["text"].encode()).hexdigest()
            elif "image" in segment["content"]:
                content_hash = hashlib.sha256(segment["content"]["image"].encode()).hexdigest()
            else:
                content_hash = str(uuid.uuid4())
        else:
            content_hash = str(uuid.uuid4())
    
        return content_hash        

    def new_segment(
        self,
        bbox,
        page=None,
        text=None,
        image=None,
        comment=None,
        page_number=None,
        page_width=None,
        page_height=None,
        rects=None,
        id=None,
    ):

        if page_width is None:
            page_width = page.rect.width
        if page_height is None:
            page_height = page.rect.height
        if page_number is None:
            page_number = page.number

        if "x1" not in bbox:
            bounding_box = self.convert_coordinates_from_pymupdf_to_pdfjs(
                bbox, page_height, page_width, page_number + 1
            )
        else:
            bounding_box = bbox

        if rects is None:
            rects = [bounding_box]

        # check if rects is of the pdfjs format
        if "x1" not in rects[0]:
            rects = [
                self.convert_coordinates_from_pymupdf_to_pdfjs(
                    rect, page_height, page_width, page_number + 1
                )
                for rect in rects
            ]

        output = {
            "content": {},
            "position": {
                "boundingRect": bounding_box,
                "rects": rects,
                "pageNumber": page_number + 1,
            },
            "comment": {"text": "", "emoji": ""},
        }

        if text is not None:
            text = text.replace("<prd>", ".")
            output["content"]["text"] = text
        if image is not None:
            output["content"]["image"] = image
        if comment is not None:
            if isinstance(comment, str):
                output["comment"]["text"] = comment
            else:
                output["comment"] = comment

        if id is None:
            output["id"] = self.generate_id(output)
        else:
            output["id"] = id
        return output

    def combine_segments(self, sA, sB):
        # take the first image
        image = None
        if "image" in sA["content"]:
            image = sA["content"]["image"]
        elif "image" in sB["content"]:
            image = sB["content"]["image"]

        # combine the text if it exists
        text = None
        if "text" in sA["content"]:
            text = sA["content"]["text"]
        if "text" in sB["content"]:
            if text is None:
                text = sB["content"]["text"]
            else:
                text += " " + sB["content"]["text"]
                # remove double spaces
                text = text.replace("  ", " ")

        # combine the comments
        comment = None
        if "comment" in sA:
            comment = sA["comment"]
        if "comment" in sB:
            if comment is None:
                comment = sB["comment"]

        # combine the bounding boxes
        bounding_boxes = [
            sA["position"]["boundingRect"],
            sB["position"]["boundingRect"],
        ]
        bounding_box = self.merge_pdfjs_bounding_boxes(bounding_boxes)

        # combine the rects
        rects = sA["position"]["rects"] + sB["position"]["rects"]

        return self.new_segment(
            bounding_box,
            text=text,
            image=image,
            comment=comment,
            page_number=sA["position"]["pageNumber"] - 1,
            page_width=sA["position"]["boundingRect"]["width"],
            page_height=sA["position"]["boundingRect"]["height"],
            rects=rects
        )


    def get_embedding(self, text, model="text-embedding-ada-002"):
        """ Get the embedding of a text using the OpenAI API.
                """
        with MongoDBConnection() as mg:
            return get_embedding(text, mg, model)


    def run(self, doc, excluded_boxes=[]):
        import pymupdf as fitz
        import requests

        if isinstance(doc, fitz.Document):
            self.file_mode = "local"
            self.file_id = doc.name
            return self._run(doc, excluded_boxes)

        elif isinstance(doc, str):
            if doc.startswith("http"):
                res = requests.get(doc, stream=True)

                self.file_mode = "remote"
                self.file_id = doc

                if res.status_code != 200:
                    raise Exception(f"Failed to download the file from {doc}")

                fitz_doc = fitz.open(stream=res.content, filetype="pdf")
            else:
                fitz_doc = fitz.open(doc)

            return self._run(fitz_doc, excluded_boxes)

        raise NotImplementedError("Subclasses must implement this method.")
    

    def save_to_cache(self, segments):
        with MongoDBConnection() as mg:
            for segment in segments:
                segment["file_id"] = self.file_id
                segment["file_mode"] = self.file_mode
                mg.segments.insert_one(segment)


    def get_from_cache(self, embedding=False):
        with MongoDBConnection() as mg:
            res = mg.segments.find(
                {"file_id": self.file_id, "file_mode": self.file_mode},
                {"_id": 0, "file_id": 0, "file_mode": 0, "embedding": 1 if embedding else 0},
            )
            return list(res)
        
    def stripped_output(self, segments, debug=False):
        stripped_segments = []
        for segment in segments:

            position = {
                "boundingRect": segment["position"]["boundingRect"],
                "rects": [] if "image" in segment["content"] else segment["position"]["rects"],
                "pageNumber": segment["position"]["pageNumber"],
            }

            stripped_segment = {
                "content": segment["content"],
                "position": position,
                "comment": segment["comment"],
                "id": segment["id"],
            }
            if debug:
                stripped_segment["comment"]["text"] = f"""ID: {segment["id"]}

{segment["content"]["text"]}"""
            stripped_segments.append(stripped_segment)
        return stripped_segments


from .base import BaseSegmenter

class SentenceSegmenter(BaseSegmenter):

    def __init__(self):
        super().__init__()

    def split_into_sentences(self, text):
        """Split into sentences using regex and keeps the sentence ending punctuation."""
        import regex as re

        # prevent splitting on decimal numbers
        text = re.sub(r"(\d)\.(\d)", r"\1<prd>\2", text)

        # fix i.e. and e.g. and etc.
        text = re.sub(r"e\.g\.", "e<prd>g<prd>", text)
        text = re.sub(r"i\.e\.", "i<prd>e<prd>", text)
        text = re.sub(r"etc\.", "etc<prd>", text)
        text = re.sub(r"vs\.", "vs<prd>", text)
        text = re.sub(r"et al\.", "et al<prd>", text)

        # if we get a name like "V. Krueckl" we don't want to split it
        text = re.sub(r"([A-Z])\.", r"\1<prd>", text)

        # special abbreviations: Fig and Eq etc e.g.
        special = [
            "Fig",
            "Eq",
            "Dr",
            "Mr",
            "Ms",
            "Mrs",
            "Prof",
            "St",
            "Ave",
            "Inc",
            "Ltd",
            "Jr",
            "Sr",
            "Co",
            "Corp",
            "Dept",
            "Univ",
            "Assn",
            "Apt",
            "Blvd",
            "Cie",
            "Cir",
            "Col",
            "Dist",
            "Fed",
            "Ft",
            "Hwy",
            "Ln",
            "Mt",
            "Op",
            "Pl",
            "Rd",
            "Sq",
            "St",
            "Tce",
            "Ter",
            "Univ",
            "Apt",
            "Dept",
            "Est",
            "Figs",
            "Eqs",
            "Dr",
            "Mr",
            "Ms",
            "Mrs",
            "Prof",
            "St",
            "Ave",
            "Inc",
            "Ltd",
            "Jr",
            "Sr",
            "Co",
            "Corp",
            "Dept",
            "Univ",
            "Assn",
            "Apt",
            "Blvd",
            "Cie",
            "Cir",
            "Col",
            "Dist",
            "Fed",
            "Ft",
            "Hwy",
            "Ln",
            "Mt",
            "Op",
            "Pl",
            "Rd",
            "Sq",
            "St",
            "Tce",
            "Ter",
            "Univ",
            "Apt",
            "Dept",
            "Est",
            "Figs",
            "Eqs",
        ]
        for el in special:
            text = re.sub(r"\b" + el + r"\.", el + "<prd>", text)

        # remove trailing whitespace and also special whitespace characters
        text = re.sub(r"\s+", " ", text.strip())

        sentence_endings = re.findall(r"[^\.!\?]+[\.!\?]*", text)
        if len(sentence_endings) == 0:
            output = [text]
        else:
            output = sentence_endings

        return output

    def ends_with_sentence_ending(self, text):
        import regex as re

        return re.match(r".*[\.\?\!]\s*$", text)
    
    def transform_special_ligature_characters(self, text):
        import unicodedata
        import re

        # replace ligature characters
        text = unicodedata.normalize("NFKD", text)
        text = re.sub(r"ﬁ", "fi", text)
        text = re.sub(r"ﬂ", "fl", text)
        text = re.sub(r"ﬀ", "ff", text)
        text = re.sub(r"ﬃ", "ffi", text)
        text = re.sub(r"ﬄ", "ffl", text)
        text = re.sub(r"ﬅ", "ft", text)
        text = re.sub(r"ﬆ", "st", text)
        
        return text
    
    def is_in_excluded_box(self, span, excluded_boxes, page_number):
        # check if the line is in an excluded box
        # excluded_boxes is in the pdfjs format
        # span uses the pymupdf format

        bbox = span["bbox"]

        for excluded_box in excluded_boxes:
            if (
                bbox[0] >= excluded_box["x1"]
                and bbox[1] <= excluded_box["x2"]
                and bbox[2] >= excluded_box["y1"]
                and bbox[3] <= excluded_box["y2"]
                and "pageNumber" in excluded_box
                and page_number == excluded_box["pageNumber"] - 1
            ):
                return True
            
        return False

    def _run(self, doc, excluded_boxes):


        sentences = []

        current_text = ""
        current_bboxes = []

        for page in doc:
            blocks = page.get_text("dict")["blocks"]

            for block in blocks:
                if "lines" in block:
                    for line in block["lines"]:
                        for span in line["spans"]:

                            # skip if the line is in an excluded box
                            if self.is_in_excluded_box(span, excluded_boxes, page.number):
                                continue

                            # get the text and apply some transformations
                            raw_text = span["text"]
                            text = self.transform_special_ligature_characters(raw_text)
                            splits = self.split_into_sentences(text)

                            # handle empty text
                            if current_text == "" or current_text == " ":
                                current_text = ""
                                current_bboxes = []

                            # loop over the splits
                            for chunk in splits:

                                if len(current_text) > 0:
                                    if current_text[-1] in ["-", "–", "—"]:
                                        current_text = current_text[:-1]
                                    else:
                                        current_text += " "
                                current_text += chunk
                                num_chunk_words = len(current_text.split())
                                current_bboxes.append(span["bbox"])

                                if (
                                    self.ends_with_sentence_ending(chunk)
                                    and len(current_text) > 6
                                    and num_chunk_words > 2
                                ):

                                    bbox = self.merge_bounding_boxes(current_bboxes)
                                    sentences.append(
                                        self.new_segment(
                                            bbox,
                                            page=page,
                                            text=current_text,
                                            rects=current_bboxes,
                                        )
                                    )
                                    current_text = ""
                                    current_bboxes = []

        return sentences



from .base import BaseSegmenter
from .figure import FigureSegmenter
from .semantic import SemanticSegmenter
from .sentence import SentenceSegmenter



class Segmenter(BaseSegmenter):

    def __init__(self,
                 chached=True,
                 calc_embedding=True,
                 max_char_joined_segments=1000,
                 max_joined_segments=50,
                 debug=False
                ):
        super().__init__()
        self.cached = chached
        self.use_embedding = calc_embedding
        self.debug = debug

        self.figure_segmenter = FigureSegmenter()
        self.sentence_segmenter = SemanticSegmenter(
            max_char_joined_segments=max_char_joined_segments,
            max_joined_segments=max_joined_segments
        )



    def _run(self, doc, excluded_boxes=[]):
        if self.cached:
            result = self.get_from_cache()
            if len(result) > 0:
                print("Returning from cache")
                return self.stripped_output(result, debug=self.debug)

        # run the figure segmenter
        self.figure_segments = self.figure_segmenter._run(doc, [])

        # run the sentence segmenter
        self.sentence_segments = self.sentence_segmenter._run(
            doc, excluded_boxes=[])

        output = self.figure_segments + self.sentence_segments



        if self.use_embedding:
            for segment in output:
                if "text" in segment["content"] and len(segment["content"]["text"]) > 2:
                    segment["embedding"] = self.get_embedding(
                        segment["content"]["text"]
                    )
            
        if self.cached:
            self.save_to_cache(output)

        
        return self.stripped_output(output, debug=self.debug)
    

def search_segments(text, filter={}, limit=10):
    from .mongo import MongoDBConnection
    from .base import get_embedding

    with MongoDBConnection() as mg:

        embedding = get_embedding(text, mg)


        return mg.vector_search(
            embedding,
            extra_filters=filter,
            limit=limit
        )


import os




class MongoDBConnection:
    def __init__(self):
        self._host = os.getenv('MONGO_HOST')
        self._db_name = os.getenv('MONGO_DB', 'pdf_highlighter')
        self.client = None
        self.db = None
        self._embedding_collection_name = "embedding_cache"
        self._segment_collection_name = "segment_cache"

        self.embeddings = None
        self.segments = None
        
        self._embedding_size = 1536
        self._embedding_name = "embedding"




    def create_vector_search_index(self, collection="segment_cache", extra_filters=["file_mode", "file_id"], name="vector_search"):

        # get the collection        
        collection = self.db[collection]

        
        # create indices for the extra filters if needed
        for el in extra_filters:
            collection.create_index([(el, "hashed")])
        
        vs_index = {
            "definition": {
                "fields": [
                    {
                        "numDimensions": self._embedding_size,
                        "path": self._embedding_name,
                        "similarity": "cosine",
                        "type": "vector"
                    },
                ],
            },
            "name": name,
            "type": "vectorSearch",
        }
        for el in extra_filters:
            vs_index["definition"]["fields"].append({
                "path": el,
                "type": "filter"
            })


        return collection.create_search_index(vs_index)

    def drop_search_index(self, collection="segment_cache", name="vector_search"):
        collection = self.db[collection]
        return collection.drop_search_index(name)
        

        
    def vector_search(self, embedding, collection=None, num_candidates=50, limit=10, extra_filters={}, name="vector_search", exact_match=False):

        if not collection:
            collection = self._segment_collection_name
       
        pipeline = [
            {
                "$vectorSearch": {
                    "exact": exact_match,
                    "index": name,
                    "path": self._embedding_name,
                    "queryVector": embedding,
                    "numCandidates": num_candidates,
                    "limit": limit,
                }
            },
            {
                "$project": {
                    "_id": 0,
                    "content": 1,
                    "position": 1,
                    "file_id": 1,
                    "file_mode": 1,
                    "search_score": { "$meta": "vectorSearchScore" }
                }
            }
        ]

        # if there are extra_filters add them to the vector_search
        if extra_filters:
            pipeline[0]["$vectorSearch"]["filter"] = extra_filters


        collection = self.db[collection]

        res = collection.aggregate(pipeline)
        return res



    def __enter__(self):
        from pymongo import MongoClient
        self.client = MongoClient(self._host)
        self.db = self.client[self._db_name]

        self.embeddings = self.db[self._embedding_collection_name]
        self.segments = self.db[self._segment_collection_name]

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.client:
            self.client.close()

# Usage example:
# with MongoDBConnection() as mg:
#     collection = mg.db['your_collection_name']
#     # Perform database operations



from __future__ import print_function as _
import os as _os


import sys as _sys
import json

import dash as _dash


# noinspection PyUnresolvedReferences
from ._imports_ import *
from ._imports_ import __all__

#from .utils import *
from .segmenter import *

if not hasattr(_dash, '__plotly_dash') and not hasattr(_dash, 'development'):
    print('Dash was not successfully imported. '
          'Make sure you don\'t have a file '
          'named \n"dash.py" in your current directory.', file=_sys.stderr)
    _sys.exit(1)

_basepath = _os.path.dirname(__file__)
_filepath = _os.path.abspath(_os.path.join(_basepath, 'package-info.json'))
with open(_filepath) as f:
    package = json.load(f)

package_name = package['name'].replace(' ', '_').replace('-', '_')
__version__ = package['version']

_current_path = _os.path.dirname(_os.path.abspath(__file__))

_this_module = _sys.modules[__name__]

_js_dist = []

_js_dist.extend(
    [
        {
            'relative_package_path': 'dash_pdf_highlighter.js',
    'external_url': 'https://unpkg.com/{0}@{2}/{1}/{1}.js'.format(
                package_name, __name__, __version__),
            'namespace': package_name
        },
        {
            'relative_package_path': 'dash_pdf_highlighter.js.map',
    'external_url': 'https://unpkg.com/{0}@{2}/{1}/{1}.js.map'.format(
                package_name, __name__, __version__),
            'namespace': package_name,
            'dynamic': True
        },
        {
            'relative_package_path': 'pdf.worker.min.mjs',
            'external_url': 'https://unpkg.com/{0}@{2}/{1}/pdf.worker.min.mjs'.format(
                package_name, __name__, __version__),
            'namespace': package_name,
            'dynamic': True
        },
        {
            'relative_package_path': '257.dash_pdf_highlighter.js',
            'external_url': 'https://unpkg.com/{0}@{2}/{1}/257.dash_pdf_highlighter.js'.format(
                package_name, __name__, __version__),
            'namespace': package_name,
            'dynamic': True
        }
    ]
)

_css_dist = []


for _component in __all__:
    setattr(locals()[_component], '_js_dist', _js_dist)
    setattr(locals()[_component], '_css_dist', _css_dist)
